# coding: utf-8

"""
Function returning the config defaults of the gfal package.
"""


def config_defaults(default_config):
    return {
        "logging": {
            "gfal2": "WARNING",
        },
    }
