# coding: utf-8
# flake8: noqa

"""
Telegram contrib functionality.
"""

__all__ = ["notify_telegram", "NotifyTelegramParameter"]


# provisioning imports
from law.contrib.telegram.notification import notify_telegram
from law.contrib.telegram.parameter import NotifyTelegramParameter
